#!/usr/bin/python3

from __future__ import annotations
from enum import Enum
from typing import Any, Optional, Type, TypeVar
import abc
import datetime
import os
import subprocess
import shlex


ItemT = TypeVar('ItemT')
TextSubT = TypeVar('TextSubT', bound='Text')
ShellCodeSubT = TypeVar('ShellCodeSubT', bound='ShellCode')


class TimeUnit(Enum):
    SECOND = 's'
    MINUTE = 'm'
    HOUR = 'h'
    DAY = 'd'
    WEEK = 'w'
    YEAR = 'y'

    def modulo(self, seconds: int) -> tuple[int, int]:
        minute = 60
        hour = minute * 60
        day = hour * 24
        week = day * 7
        year = day * 365.25
        size = {
            self.SECOND: 1,
            self.HOUR: hour,
            self.MINUTE: minute,
            self.DAY: day,
            self.WEEK: week,
            self.YEAR: year,
        }
        if seconds < minute:
            return 0, seconds
        if self == self.SECOND:
            return seconds, 0
        if self not in size:
            raise ValueError(f"unsupported unit: {self}")
        remainder = seconds % size[self]
        amount = int((seconds - remainder) / size[self])
        return amount, remainder


class _Void:
    pass


_VOID = _Void()


class ParsingError(ValueError):

    def __init__(self,
                 description: str | None = None,
                 loc: list[int | str] | None = None,
                 value: str | _Void = _VOID,
                 ):
        self.description = description
        self.loc = loc or []
        self.value = value


def sequence_duration(seconds: int) -> dict[TimeUnit, int]:
    """
    Split *seconds* duration to larger units, up to weeks

    Note that we avoid months like the plague.
    """
    rest = seconds
    parsed: dict[TimeUnit, int] = {}
    parsed[TimeUnit.YEAR], rest = TimeUnit.YEAR.modulo(rest)
    parsed[TimeUnit.WEEK], rest = TimeUnit.WEEK.modulo(rest)
    parsed[TimeUnit.DAY], rest = TimeUnit.DAY.modulo(rest)
    parsed[TimeUnit.HOUR], rest = TimeUnit.HOUR.modulo(rest)
    parsed[TimeUnit.MINUTE], rest = TimeUnit.MINUTE.modulo(rest)
    parsed[TimeUnit.SECOND] = rest
    return parsed


def fmt_duration(seconds: int) -> str:
    """
    Translate to a valid duration specification

    keep up to 2 largest non-zero units, but at least
    seconds if all else is zeroed
    """
    sequenced = sequence_duration(seconds)
    units = [TimeUnit.YEAR, TimeUnit.WEEK, TimeUnit.DAY, TimeUnit.HOUR, TimeUnit.MINUTE, TimeUnit.SECOND]
    parts: list[str] = []
    for unit in units:
        if not sequenced[unit]:
            continue
        parts.append(str(sequenced[unit]) + unit.value)
    return ','.join(parts)


def parse_duration_brief(expressions: list[str]) -> float:
    """
    Parse time duration text to number of seconds

    Each string in *expressions* must be in format:

        <NUMBER><UNIT>

    where NUMBER must be a number no lower than zero, and UNIT
    can be one of 's', 'm', 'h', 'd', 'w' or 'y'.  For example,
    '15m' is a valid EXPRESSION representing fifteen minutes.

    Individual expressions add up, so, eg. `["1h", "3m"]` will
    result in seconds corresponding to 63 minutes.

    UNIT represents 's' for second, 'm' for minute, 'h' for hourm
    'd' for day, 'w' for week and 'y' for year.  Months are not
    supported.  (They are evil, in case you haven't noticed.)
    """

    minute = 60
    hour = minute * 60
    day = hour * 24
    week = day * 7
    year = day * 365.25

    def parse_part(P):
        if P.endswith('s'):
            return float(P[:-1])
        elif P.endswith('m'):
            return float(P[:-1]) * minute
        elif P.endswith('h'):
            return float(P[:-1]) * hour
        elif P.endswith('d'):
            return float(P[:-1]) * day
        elif P.endswith('w'):
            return float(P[:-1]) * week
        elif P.endswith('y'):
            return float(P[:-1]) * year
        else:
            return float(P)
    total = 0
    for idx, part in enumerate(expressions):
        try:
            total += parse_part(part)
        except ValueError:
            raise ParsingError(value=part, loc=[idx])
    return total


class BashSyntaxError(ValueError):
    pass


class CommandError(RuntimeError):
    pass


def nlstrip(text: str) -> str:
    if text.endswith('\n'):
        return text[0:-1]
    return text


def nlsplit(text: str) -> list[str]:
    if not text:
        return []
    return nlstrip(text).split('\n')


def nljoin(lines: list[str]) -> str:
    return ''.join(line + '\n' for line in lines)


class Text:

    @classmethod
    def new(cls: Type[TextSubT], value: Any = None) -> TextSubT:
        if value is None:
            return cls([], _fail=False)
        if type(value) is str:
            return cls.from_str(value)
        if type(value) is bytes:
            return cls.from_bytes(value)
        if type(value) is list:
            return cls(lines=value, _fail=False)
        if type(value) is cls:
            return cls(lines=value.as_lines, _fail=False)
        raise ValueError(
            f"value must be {cls.__name__}, str, bytes, list or None"
        )

    @classmethod
    def from_bytes(cls: Type[TextSubT], value: Any = None) -> TextSubT:
        if value is None:
            return cls.new()
        return cls.from_str(value.decode())

    @classmethod
    def from_lines(cls: Type[TextSubT], value: Any = None) -> TextSubT:
        if value is None:
            return cls.new()
        return cls(value, _fail=False)

    @classmethod
    def from_str(cls: Type[TextSubT], value: Any = None) -> TextSubT:
        if value is None:
            return cls.new()
        return cls(nlsplit(value), _fail=False)

    def __init__(self, lines: list[str] = [], _fail: bool = True):
        if _fail:
            cls = self.__class__
            raise ValueError(
                f"don't instantiate {cls.__name__} directly;"
                f"use {cls.__name__}.from_*() or {cls.__name__}.new()"
            )
        self._lines = lines or []

    def __repr__(self):
        return f'{self.__class__.__name__}({self._lines!r})'

    def __str__(self):
        return self.as_str

    def prefix(self, pfx: str) -> Text:
        return self.__class__.from_lines([
            pfx + line for line in self._lines
        ])

    @property
    def as_bytes(self) -> bytes:
        return self.as_str.encode()

    @property
    def as_lines(self) -> list[str]:
        return list(line for line in self._lines)

    @property
    def as_str(self) -> str:
        return nljoin(self._lines)


class TextFile:
    """
    Text file working with Text() objects
    """

    def __init__(self, path: str):
        self.path = path
        self.base, self.leaf = path.rsplit('/', maxsplit=1)

    def _ensure_dir(self) -> None:
        if os.path.isdir(self.base):
            return
        os.makedirs(self.base)

    def exists(self) -> bool:
        return os.path.isfile(self.path)

    def append(self, content: Any) -> None:
        txt = Text.new(content)
        self._ensure_dir()
        with open(self.path, 'a') as fh:
            fh.write(txt.as_str)

    def load(self) -> Text:
        if not self.exists():
            return Text.new()
        with open(self.path) as fh:
            return Text.from_str(fh.read())

    def save(self, content: Any) -> None:
        txt = Text.new(content)
        self._ensure_dir()
        with open(self.path, 'w') as fh:
            fh.write(txt.as_str)


class TimePeriod:

    @classmethod
    def after(cls, start: datetime.datetime) -> TimePeriod:
        return cls(
            start=start,
            end=datetime.datetime.now()
        )

    def __init__(self,
                 start: datetime.datetime,
                 end: datetime.datetime,
                 ):
        self.start = start
        self.end = end

    @property
    def start_ts(self) -> str:
        return self.start.isoformat()

    @property
    def end_ts(self) -> str:
        return self.end.isoformat()


class CmdResult:

    @classmethod
    def new(cls,
            cmd: Cmd,
            out: bytes,
            err: bytes,
            es: int,
            timing: TimePeriod):
        return cls(
            cmd=cmd,
            out=Text.new(out),
            err=Text.new(err),
            es=es,
            timing=timing,
        )

    def __init__(self,
                 cmd: Cmd,
                 out: Text,
                 err: Text,
                 es: int,
                 timing: TimePeriod):
        self.cmd = cmd
        self.out = out
        self.err = err
        self.es = es
        self.timing = timing

    def fmt(self) -> Text:
        def lines():
            yield f'command exited with: {self.es}'
            yield f' .. start: {self.timing.start_ts}'
            yield f' .. end:   {self.timing.end_ts}'
            for line in self.cmd.spec.code.as_lines:
                yield f' .. code:|{line}'
            for line in self.err.as_lines:
                yield f' .. stderr:|{line}'
            len_c = len(self.out.as_str)
            len_l = len(self.out.as_str)
            yield f' .. stdout: (not showing {len_c} chars / {len_l} lines)'
        return Text.from_lines(lines()).prefix('{%s}: ' % self.cmd.role)


class ShellCode(Text, metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def as_cmd(self) -> list[str]:
        pass

    @property
    @abc.abstractmethod
    def syn_exc(self) -> Type:
        pass

    @classmethod
    def from_str(cls: Type[ShellCodeSubT],
                 value: Any = None,
                 env: dict[str, str] = {}
                 ) -> ShellCodeSubT:
        if value is None:
            return cls.new()
        env_lines = [f'export {k}={shlex.quote(v)}' for k, v in env.items()]
        return cls(env_lines + nlsplit(value), _fail=False)

    def check_syntax(self) -> None:
        check = Cmd(
            spec=CmdSpec.from_args(['bash', '-n']),
            stdin=self,
            role='bmo.util.ShellCode.check_syntax()',
        )
        result = check.run()
        if result.es > 0:
            txt = result.err.prefix('err:|')
            raise self.syn_exc(txt.as_str)


class BashCode(ShellCode):

    @property
    def as_cmd(self) -> list[str]:
        return ['bash', '-c', self.as_str]

    @property
    def syn_exc(self) -> Type:
        return BashSyntaxError


class CmdSpec:

    @classmethod
    def from_args(cls, args: list[str]) -> CmdSpec:
        return cls(args)

    @classmethod
    def from_code(cls, code: ShellCode) -> CmdSpec:
        code.check_syntax()
        return cls(args=code.as_cmd, code=code)

    def __init__(self,
                 args: list[str],
                 code: ShellCode | None = None,
                 ):
        self.args = args
        self._code = code

    @property
    def code(self) -> ShellCode:
        if self._code is None:
            return ShellCode.from_str(str(self.args))
        return self._code


class Cmd:

    @classmethod
    def from_code(cls,
                  code: ShellCode,
                  stdin: str | None = None,
                  role: str | None = None,
                  ):
        """
        Construct command from Bash *code*
        """
        return cls(
            spec=CmdSpec.from_code(code),
            stdin=Text.new(stdin),
            role=role,
        )

    def __init__(self,
                 spec: CmdSpec,
                 stdin: Text | None = None,
                 role: str | None = None,
                 ):
        self.spec = spec
        self.stdin = stdin or Text.new()
        self.role = role or '(some command)'
        self._ts_start: Optional[datetime.datetime] = None

    def run(self) -> CmdResult:
        self._ts_start = datetime.datetime.now()
        pipe = subprocess.Popen(
            self.spec.args,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = pipe.communicate(input=self.stdin.as_bytes)
        return CmdResult.new(
            cmd=self,
            out=out,
            err=err,
            es=pipe.returncode,
            timing=TimePeriod.after(self._ts_start),
        )
